/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file checksum.c
 * Implementation of the s3 DSI checksum
 */
#include "globus_gridftp_server_s3_base.h"

static
void *
globus_l_s3_checksum_thread(
    void                               *user_arg);

static
S3Status
globus_l_s3_checksum_data(
    int                                 bufferSize,
    const char                         *buffer,
    void                               *callbackData);

static
S3Status
globus_l_s3_checksum_compute_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
S3Status
globus_l_s3_checksum_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);


static
void
globus_l_s3_checksum_complete(
    S3Status                            status,
    const S3ErrorDetails               *errorDetails,
    void                               *callbackData);

typedef struct
{
    S3BucketContext                     bucket_context;
    char                               *bucket_name;
    char                               *key;
    globus_gfs_operation_t              op;
    bool                                send_marker;
    globus_off_t                        offset;
    globus_off_t                        length;
    globus_off_t                        processed;
    int                                 retries;
    S3Status                            status;
    globus_result_t                     result;
    int                                 marker_update_interval;
    time_t                              next_update_marker_time;

    MD5_CTX                             checksum_context;
    char                                checksum[MD5_DIGEST_LENGTH*2+1];
    char                                pathname[];
}
globus_l_s3_checksum_t;

/**
 * @brief checksum
 * @details
 */
void
globus_i_s3_checksum(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_l_s3_checksum_t             *checksum_state = NULL;
    globus_thread_t                     t;
    int                                 rc;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    assert(s3_handle != NULL);
    assert(cmd_info != NULL);
    assert(cmd_info->pathname != NULL);
    assert(cmd_info->cksm_alg != NULL);

    GlobusS3Info("checksum: op=%p "
                 "pathname=%s "
                 "alg=%s "
                 "offset=%"PRIdMAX" "
                 "length=%"PRIdMAX "\n",
                 (void *) op,
                 cmd_info->pathname,
                 cmd_info->cksm_alg,
                 (intmax_t) cmd_info->cksm_offset,
                 (intmax_t) cmd_info->cksm_length);

    if ((cmd_info->cksm_alg == NULL)
        || strcasecmp(cmd_info->cksm_alg, "md5") != 0)
    {
        const char                      msg_fmt[] = "Unknown checksum algorithm \"%s\"";
        size_t                          msg_len = snprintf(NULL, 0, msg_fmt, cmd_info->cksm_alg);
        char                            msg[msg_len+1];

        snprintf(msg, sizeof(msg), msg_fmt, cmd_info->cksm_alg);

        result = GlobusGFSErrorGeneric(msg);

        goto fail;
    }

    checksum_state = malloc(sizeof(globus_l_s3_checksum_t) + strlen(cmd_info->pathname) + 1);
    if (checksum_state == NULL)
    {
        result = GlobusGFSErrorMemory("checksum");

        goto fail;
    }

    *checksum_state = (globus_l_s3_checksum_t)
    {
        .op                           = op,
        .offset                       = cmd_info->cksm_offset,
        .length                       = cmd_info->cksm_length
    };
    globus_gridftp_server_get_update_interval(op,
            &checksum_state->marker_update_interval);

    strcpy(checksum_state->pathname, cmd_info->pathname);

    globus_i_s3_path_parse(checksum_state->pathname, &checksum_state->bucket_name, &checksum_state->key);
    checksum_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, checksum_state->bucket_name);

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_checksum_thread,
            checksum_state);


    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorSystemError("globus_thread_create", errno);

fail:
        free(checksum_state);
        GlobusS3Info("finished_command: op=%p result=%#x\n",
                    (void *) op,
                    (unsigned int) result);
        globus_gridftp_server_finished_command(op, result, NULL);
    }

    GlobusS3Exit();

    return;
}
/* globus_i_s3_checksum() */

static
void *
globus_l_s3_checksum_thread(
    void                               *user_arg)
{
    globus_l_s3_checksum_t             *checksum_state = user_arg;

    do
    {
        checksum_state->status = S3StatusOK;
        checksum_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(checksum_state->retries);

        GlobusS3Info("REST: op=%p S3_head_object(bucket=%s, key=%s) try=%d\n",
                    (void *) checksum_state->op,
                    checksum_state->bucket_name,
                    checksum_state->key,
                    checksum_state->retries);

        S3_head_object(
                &checksum_state->bucket_context,
                checksum_state->key,
                NULL,
                &(S3ResponseHandler)
                {
                    .propertiesCallback   = globus_l_s3_checksum_properties,
                    .completeCallback     = globus_l_s3_checksum_complete,
                },
                checksum_state);

        GlobusS3DebugOpStatus(S3_head_object, checksum_state);
    }
    while (S3_status_is_retryable(checksum_state->status)
            && (++checksum_state->retries) < globus_i_s3_max_retries);

    if (checksum_state->result != GLOBUS_SUCCESS)
    {
        goto bad_head;
    }

    if (checksum_state->checksum[0] == 0)
    {
        int rc;

        /* No checksum (file too big and no metadata). Fetch the data from
         * the object store and compute the checksum
         */
        GlobusS3Info("op=%p no_md5_metadata_available\n",
                (void *) checksum_state->op);

        rc = MD5_Init(&checksum_state->checksum_context);
        if (rc != 1)
        {
            checksum_state->result = GlobusGFSErrorMemory("checksum_context");

            goto md5_init_fail;
        }
        /* Assume posix semantics for time */
        checksum_state->next_update_marker_time = time(NULL)
                + checksum_state->marker_update_interval;

        checksum_state->retries = 0;
        do
        {
            checksum_state->status = S3StatusOK;
            checksum_state->result = GLOBUS_SUCCESS;

            globus_i_s3_retry_delay(checksum_state->retries);

            GlobusS3Info("REST: op=%p S3_get_object(bucket=%s, key=%s) try=%d\n",
                        (void *) checksum_state->op,
                        checksum_state->bucket_name,
                        checksum_state->key,
                        checksum_state->retries);
            S3_get_object(
                    &checksum_state->bucket_context,
                    checksum_state->key,
                    NULL,
                    checksum_state->offset,
                    (checksum_state->length == (globus_off_t) -1)
                            ? 0 : checksum_state->length,
                    NULL,
                    &(S3GetObjectHandler)
                    {
                        .responseHandler  = 
                        {
                            .propertiesCallback
                                          = globus_l_s3_checksum_compute_properties,
                            .completeCallback
                                          = globus_l_s3_checksum_complete,
                        },
                        .getObjectDataCallback
                                          = globus_l_s3_checksum_data
                    },
                    checksum_state);
            GlobusS3DebugOpStatus(S3_get_object, checksum_state);
        }
        while (S3_status_is_retryable(checksum_state->status)
            && (checksum_state->retries++) < globus_i_s3_max_retries);

        if (checksum_state->result == GLOBUS_SUCCESS)
        {
            unsigned char                   checksum[MD5_DIGEST_LENGTH];

            MD5_Final(checksum, &checksum_state->checksum_context);

            for (size_t i = 0; i < MD5_DIGEST_LENGTH; i++)
            {
                sprintf(&checksum_state->checksum[2*i], "%.2x", (unsigned int) checksum[i]);
            }
        }
    }

md5_init_fail:
bad_head:

    GlobusS3Info("finished_command: op=%p result=%#x checksum=%.*s\n",
                (void *) checksum_state->op,
                (unsigned int) checksum_state->result,
                (int) sizeof(checksum_state->checksum),
                checksum_state->checksum);

    globus_gridftp_server_finished_command( 
            checksum_state->op,
            checksum_state->result,
            checksum_state->checksum[0] ? checksum_state->checksum : NULL);

    free(checksum_state);

    return NULL;
}
/* globus_l_s3_checksum_thread() */

static
S3Status
globus_l_s3_checksum_data(
    int                                 buffer_size,
    const char                         *buffer,
    void                               *callback_data)
{
    int                                 rc;
    S3Status                            status = S3StatusOK;
    globus_l_s3_checksum_t             *checksum_state = callback_data;

    GlobusS3Enter();

    rc = MD5_Update(&checksum_state->checksum_context,
               buffer,
               (unsigned long long) buffer_size);
    if (rc != 1)
    {
        status = S3StatusInternalError;

        checksum_state->result = GlobusGFSErrorGeneric("MD5_Update error");
    }
    checksum_state->processed += buffer_size;

    if (checksum_state->marker_update_interval != 0)
    {
        time_t                          now = time(NULL);

        /* Assume posix semantics for time */
        if (now >= checksum_state->next_update_marker_time)
        {
            const char *fmt = "%"GLOBUS_OFF_T_FORMAT;
            size_t len = snprintf(NULL, 0, fmt, checksum_state->processed);
            char marker[len+1];

            snprintf(marker, sizeof(marker), fmt, checksum_state->processed);
            globus_gridftp_server_intermediate_command(
                    checksum_state->op, GLOBUS_SUCCESS, marker);

            checksum_state->next_update_marker_time =
                    now + checksum_state->marker_update_interval;
        }
    }

    GlobusS3ExitStatus(status);
    return status;
}
/* globus_l_s3_checksum_data() */

static
S3Status
globus_l_s3_checksum_compute_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
static
S3Status
globus_l_s3_checksum_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    globus_l_s3_checksum_t             *checksum_state = callback_arg;
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3Trace("requestId=%s requestId2=%s contentType=%s contentLength=%"PRIu64" server=%s eTag=%s lastModified=%"PRId64"\n",
            properties->requestId ? properties->requestId : "\"\"",
            properties->requestId2 ? properties->requestId2 : "\"\"",
            properties->contentType ? properties->contentType : "\"\"",
            properties->contentLength,
            properties->server ? properties->server : "\"\"",
            properties->eTag ? properties->eTag : "\"\"",
            properties->lastModified);

    if ((checksum_state->offset == 0) &&
        ((checksum_state->length == properties->contentLength)
          || checksum_state->length == (globus_off_t) -1))
    {
        int                             rc;

        if ((properties->contentLength < UINT64_C(5000000000)) && properties->eTag && strlen(properties->eTag) == 34)
        {
            GlobusS3Trace("Setting md5sum=%.32s from etag\n", properties->eTag+1);

            rc = sscanf(properties->eTag, "\"%32[a-fA-F0-9]\"", checksum_state->checksum);
            if (rc != 1 || strlen(checksum_state->checksum) != 32)
            {
                /* Ignore invalid checksum etag */
                checksum_state->checksum[0] = 0;
            }
        }
        /* Allow metadata set at object creation to override etag */
        for (int i = 0; i < properties->metaDataCount; i++)
        {
            if (strcmp(properties->metaData[i].name, GLOBUS_S3_MD5_NAME) == 0)
            {
                if (strlen(properties->metaData[i].value) == 32)
                {
                    GlobusS3Trace("Setting md5sum=%s from metadata\n",
                            properties->metaData[i].value);

                    rc = sscanf(properties->metaData[i].value, "%32[a-fA-F0-9]", checksum_state->checksum);
                    if (rc != 1 || strlen(checksum_state->checksum) != 32)
                    {
                        /* Ignore invalid checksum etag */
                        checksum_state->checksum[0] = 0;
                    }
                    break;
                }
            }
        }
    }

    GlobusS3ExitStatus(status);
    return status;
}
/* globus_l_s3_checksum_properties() */

static
void
globus_l_s3_checksum_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_l_s3_checksum_t             *checksum_state = callback_data;
    GlobusS3Enter();

    if (status != S3StatusOK)
    {
        char path[strlen(checksum_state->bucket_name) + strlen(checksum_state->key) + 2];
        snprintf(path, sizeof(path), "%s/%s", checksum_state->bucket_name,checksum_state->key);

        checksum_state->status = status;
        checksum_state->result = GlobusGFSErrorS3(
                    status,
                    path,
                    error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_checksum_complete() */
// vim: ft=c:
