/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_gridftp_server_s3_stat.h"

#include <libgen.h>

static
S3Status
globus_l_s3_stat_head_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
void
globus_l_s3_stat_head_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_stat_list_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
void
globus_l_s3_stat_list_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_stat_list_dir_callback(
    int                                 is_truncated,
    const char                         *next_marker,
    int                                 contents_count, 
    const S3ListBucketContent          *contents,
    int                                 common_prefixes_count,
    const char                        **common_prefixes,
    void                               *callback_data);

static
S3Status
globus_l_s3_stat_head_object_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

static
void
globus_l_s3_stat_head_object_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

void
globus_i_s3_stat_list_objects(
    globus_i_s3_stat_t                 *stat_state)
{
    char                                keydir[strlen(stat_state->key?stat_state->key:"")+2];

    GlobusS3Enter();

    sprintf(keydir, "%s/", stat_state->key?stat_state->key:"");

    stat_state->retries = 0;

    do
    {
        stat_state->status = S3StatusOK;
        stat_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(stat_state->retries);

        /* Head "key/" to see if we created a directory placeholder */
        GlobusS3Info("REST: op=%p S3_head_object(bucket=%s, key=%s) try=%d\n",
                    (void *) stat_state->op,
                    stat_state->bucket_name,
                    keydir,
                    stat_state->retries);
        S3_head_object(
                &stat_state->bucket_context,
                keydir,
                NULL,
                &(S3ResponseHandler)
                {
                    .propertiesCallback
                                      = globus_l_s3_stat_head_dir_properties,
                    .completeCallback = globus_l_s3_stat_head_dir_complete
                },
                stat_state);

        GlobusS3DebugOpStatus(S3_head_object, stat_state);
    }
    while (S3_status_is_retryable(stat_state->status)
        && (stat_state->retries++) < globus_i_s3_max_retries);

    if (stat_state->result != GLOBUS_SUCCESS)
    {
        return;
    }

    /* List "key/" to see if it's the prefix to a key */
    do
    {
        char *next_marker = stat_state->next_marker;
        stat_state->next_marker = NULL;

        stat_state->retries = 0;

        do
        {
            stat_state->status = S3StatusOK;
            stat_state->result = GLOBUS_SUCCESS;

            globus_i_s3_retry_delay(stat_state->retries);

            GlobusS3Info("REST: op=%p S3_list_bucket(bucket=%s, key=%s) try=%d\n",
                        (void *) stat_state->op,
                        stat_state->bucket_name,
                        stat_state->key ? keydir : "<NULL>",
                        stat_state->retries);
            S3_list_bucket(
                    &stat_state->bucket_context,
                    stat_state->key ? keydir : NULL,
                    next_marker,
                    "/",
                    0,
                    NULL,
                    &(S3ListBucketHandler)
                    {
                        .responseHandler = (S3ResponseHandler)
                        {
                            .propertiesCallback
                                          = globus_l_s3_stat_list_dir_properties,
                            .completeCallback
                                          = globus_l_s3_stat_list_dir_complete
                        },
                        .listBucketCallback
                                          = globus_l_s3_stat_list_dir_callback
                    },
                    stat_state);

            GlobusS3DebugOpStatus(S3_list_bucket, stat_state);
        }
        while (S3_status_is_retryable(stat_state->status)
                && (stat_state->retries++) < globus_i_s3_max_retries);
        free(next_marker);
    } while (stat_state->result == GLOBUS_SUCCESS && stat_state->next_marker);

    if (stat_state->result != GLOBUS_SUCCESS || stat_state->is_dir)
    {
        /* If this is a directory stat, we've completed listing the contents
         * now
         */
        return;
    }

    stat_state->retries = 0;

    /* Head key to see if it's a file */
    do
    {
        stat_state->status = S3StatusOK;
        stat_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(stat_state->retries);

        GlobusS3Info("REST: op=%p S3_head_object(bucket=%s, key=%s) try=%d\n",
                    (void *) stat_state->op,
                    stat_state->bucket_name,
                    stat_state->key,
                    stat_state->retries);
        S3_head_object(
                &stat_state->bucket_context,
                stat_state->key,
                NULL,
                &(S3ResponseHandler)
                {
                    .propertiesCallback   = globus_l_s3_stat_head_object_properties,
                    .completeCallback     = globus_l_s3_stat_head_object_complete
                },
                stat_state);
        GlobusS3DebugOpStatus(S3_head_object, stat_state);
    }
    while (S3_status_is_retryable(stat_state->status)
            && (stat_state->retries++) < globus_i_s3_max_retries);

    GlobusS3Exit();
}
/* globus_l_s3_stat_list_objects() */

static
S3Status
globus_l_s3_stat_head_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;
    globus_i_s3_stat_t                 *stat_state = callback_arg;


    GlobusS3Enter();


    if (properties != NULL
        && properties->contentType != NULL
        && strcmp(properties->contentType, GLOBUS_S3_DIR_PLACEHOLDER_TYPE) == 0)
    {
        char                                dirname[strlen(stat_state->stat_info->pathname) + 2];
        int                                 dev, ino;

        snprintf(
                dirname,
                sizeof(dirname),
                "%s/",
                stat_state->stat_info->pathname);

        dev = globus_hashtable_string_hash(stat_state->bucket_name, INT_MAX);
        ino = globus_hashtable_string_hash(dirname, INT_MAX);

        stat_state->is_dir = true;

        globus_i_s3_stat_add(
                stat_state,
                &(globus_gfs_stat_t)
                {
                    .mode             = S_IRWXU|S_IFDIR,
                    .nlink            = 1,
                    .name             = ".",
                    .dev              = dev,
                    .ino              = ino,
                    .mtime            = (time_t) -1,
                    .atime            = (time_t) -1,
                    .ctime            = (time_t) -1
                });
    }


    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_head_dir_properties() */

static
void
globus_l_s3_stat_head_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_stat_t                 *stat_state = callback_data;

    GlobusS3Enter();

    if (status == S3StatusOK)
    {
        stat_state->is_dir = true;
    }
    else if (status != S3StatusOK && status != S3StatusHttpErrorNotFound)
    {
        stat_state->status = status;
        stat_state->result = 
                GlobusGFSErrorS3(
                    status,
                    stat_state->stat_info->pathname,
                    error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_stat_head_dir_complete() */

static
S3Status
globus_l_s3_stat_list_dir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_list_dir_properties() */

static
void
globus_l_s3_stat_list_dir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_stat_t                 *stat_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && status != S3StatusHttpErrorNotFound && stat_state->result == GLOBUS_SUCCESS)
    {
        stat_state->status = status;
        stat_state->result = 
                GlobusGFSErrorS3(
                    status,
                    stat_state->stat_info->pathname,
                    error_details);
    }

    GlobusS3Exit();
}
/* globus_l_s3_stat_list_dir_complete() */

static
S3Status
globus_l_s3_stat_list_dir_callback(
    int                                 is_truncated,
    const char                         *next_marker,
    int                                 contents_count, 
    const S3ListBucketContent          *contents,
    int                                 common_prefixes_count,
    const char                        **common_prefixes,
    void                               *callback_data)
{

    globus_i_s3_stat_t                 *stat_state = callback_data;
    S3Status                            status = S3StatusOK;
    char                                dirname[strlen(stat_state->key?stat_state->key:"") + 2];
    int                                 dev, ino;


    GlobusS3Enter();

    if (stat_state->key != NULL)
    {
        snprintf(dirname, sizeof(dirname), "%s/", stat_state->key);
    }
    else
    {
        dirname[0] = 0;
    }

    for (size_t i = 0; i < (size_t) contents_count; i++)
    {
        if (strncmp(contents[i].key, dirname,strlen(dirname)) == 0)
        {
            if (stat_state->is_dir == false)
            {
                stat_state->is_dir = true;
                dev = globus_hashtable_string_hash(
                        (void *) stat_state->bucket_name,
                        INT_MAX);
                ino = globus_hashtable_string_hash(
                        (void *) (dirname[0] ? dirname : "/"),
                        INT_MAX);

                globus_i_s3_stat_add(
                        stat_state,
                        &(globus_gfs_stat_t)
                        {
                            .mode             = S_IRWXU|S_IFDIR,
                            .nlink            = 1,
                            .name             = ".",
                            .dev              = dev,
                            .ino              = ino,
                            .mtime            = (time_t) -1,
                            .atime            = (time_t) -1,
                            .ctime            = (time_t) -1
                        });
            }
            if (strcmp(contents[i].key, dirname) != 0)
            {
                char keyname[strlen(contents[i].key)+1];
                char *base = NULL;

                snprintf(keyname, sizeof(keyname), "%s", contents[i].key);

                base = basename(keyname);

                dev = globus_hashtable_string_hash(
                        (void *) stat_state->bucket_name,
                        INT_MAX);
                ino = globus_hashtable_string_hash(
                        (void *) contents[i].key,
                        INT_MAX);

                globus_i_s3_stat_add(
                        stat_state,
                        &(globus_gfs_stat_t)
                        {
                            .mode             = S_IRWXU|S_IFREG,
                            .nlink            = 1,
                            .name             = base,
                            .dev              = dev,
                            .ino              = ino,
                            .size             = contents[i].size,
                            .mtime            = (contents[i].lastModified >= 0)
                                                ? contents[i].lastModified
                                                : (time_t) -1,
                            .atime            = (time_t) -1,
                            .ctime            = (time_t) -1
                        });
            }
        }
    }
    for (size_t i = 0; i < (size_t) common_prefixes_count; i++)
    {
        if (strncmp(common_prefixes[i], dirname, strlen(dirname)) == 0)
        {
            if (stat_state->is_dir == false)
            {
                stat_state->is_dir = true;
                dev = globus_hashtable_string_hash(
                        (void *) stat_state->bucket_name,
                        INT_MAX);
                ino = globus_hashtable_string_hash(
                        (void *) (dirname[0] ? dirname : "/"),
                        INT_MAX);

                globus_i_s3_stat_add(
                        stat_state,
                        &(globus_gfs_stat_t)
                        {
                            .mode             = S_IRWXU|S_IFDIR,
                            .nlink            = 1,
                            .name             = ".",
                            .dev              = dev,
                            .ino              = ino,
                            .mtime            = (time_t) -1,
                            .atime            = (time_t) -1,
                            .ctime            = (time_t) -1
                        });
            }
            if (strcmp(common_prefixes[i], dirname) != 0)
            {
                char keyname[strlen(common_prefixes[i])+1];
                char *base = NULL;

                snprintf(keyname, sizeof(keyname), "%s", common_prefixes[i]);

                base = basename(keyname);

                dev = globus_hashtable_string_hash(
                        (void *) stat_state->bucket_name,
                        INT_MAX);
                ino = globus_hashtable_string_hash(
                        (void *) common_prefixes[i],
                        INT_MAX);

                globus_i_s3_stat_add(
                        stat_state,
                        &(globus_gfs_stat_t)
                        {
                            .mode             = S_IRWXU|S_IFDIR,
                            .nlink            = 1,
                            .name             = base,
                            .dev              = dev,
                            .ino              = ino,
                            .mtime            = (time_t) -1,
                            .atime            = (time_t) -1,
                            .ctime            = (time_t) -1
                        });
            }
        }
    }
    if (is_truncated && next_marker)
    {
        stat_state->next_marker = strdup(next_marker);
        if (stat_state->next_marker == NULL)
        {
            stat_state->result = GlobusGFSErrorMemory("next_marker");
            status = S3StatusOutOfMemory;
        }
    }
    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_list_dir_callback() */

static
S3Status
globus_l_s3_stat_head_object_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    globus_i_s3_stat_t                 *stat_state = callback_arg;
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    if (properties != NULL)
    {
        int                                 dev, ino;
        mode_t                              mode;
        char pathname[strlen(stat_state->key?stat_state->key:".")+1];
        char *base = NULL;

        if (stat_state->key == NULL)
        {
            /* Don't send a stat result of bucket name here if we've already
             * handle "."
             */
            if (stat_state->is_dir)
            {
                goto done;
            }
            stat_state->is_dir = true;
            mode = S_IFDIR|S_IRWXU;
            snprintf(pathname, sizeof(pathname), "%s", ".");
        }
        else
        {
            mode = S_IFREG|S_IRWXU;
            snprintf(pathname, sizeof(pathname), "%s", stat_state->key);
        }

        base = basename(pathname);

        dev = globus_hashtable_string_hash(stat_state->bucket_name, INT_MAX);
        ino = globus_hashtable_string_hash(stat_state->stat_info->pathname, INT_MAX);

        globus_i_s3_stat_add(
                stat_state,
                &(globus_gfs_stat_t)
                {
                    .mode             = mode,
                    .nlink            = 1,
                    .name             = base,
                    .dev              = dev,
                    .ino              = ino,
                    .size             = (mode&S_IFREG) ? properties->contentLength : 0,
                    .mtime            = (mode&S_IFREG) ? (time_t) properties->lastModified : (time_t) -1,
                    .atime            = (time_t) -1,
                    .ctime            = (time_t) -1
                });
    }
done:

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_stat_head_object_properties() */

static
void
globus_l_s3_stat_head_object_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_stat_t                 *stat_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && stat_state->result == GLOBUS_SUCCESS)
    {
        stat_state->status = status;
        stat_state->result = GlobusGFSErrorS3(status, stat_state->stat_info->pathname, error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_stat_head_object_complete() */

// vim: ft=c:
